use crate::user::SaltedPassword;
use rust_decimal::Decimal;
use sodiumoxide::crypto::pwhash::argon2id13::{pwhash_verify, HashedPassword};
use sqlx::{query, PgPool};
use std::convert::TryFrom;
use thiserror::Error;
use vtse_common::{
    net::{ServerResponse, UserError},
    user::User,
};
use vtse_common::{stock::Stock, user::UserBalance};
use vtse_common::{
    stock::StockSymbol,
    user::{ApiKey, Password, UserDebt, Username},
};

#[derive(Error, Debug)]
pub(crate) enum StateError {
    #[error("Was not in the correct state")]
    WrongState,
    #[error("Got SQLx error: {0}`")]
    Database(#[from] sqlx::Error),
    #[error("Failed to hash password")]
    PasswordHash,
}

type OperationResult = Result<ServerResponse, StateError>;

#[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Hash, Debug)]
pub(crate) enum AppState {
    Unauthorized,
    Authorized { user_id: i32 },
}

/// Helper functions
impl AppState {
    pub(crate) fn new() -> Self {
        Self::Unauthorized
    }

    fn assert_state(&self, expected_state: AppState) -> Result<(), StateError> {
        if *self != expected_state {
            Err(StateError::WrongState)
        } else {
            Ok(())
        }
    }
}

/// Query operations impl
impl AppState {
    pub(crate) async fn stock_info(
        &self,
        stock_symbol: StockSymbol,
        pool: &PgPool,
    ) -> OperationResult {
        let stock = query!(
            "SELECT name, symbol, description, price FROM stocks WHERE symbol = $1",
            stock_symbol.inner()
        )
        .fetch_one(pool)
        .await?;

        Ok(ServerResponse::StockInfo(Stock::new(
            stock.name,
            stock.symbol,
            stock.description,
            stock.price,
        )))
    }

    pub(crate) async fn user_info(&self, username: Username, pool: &PgPool) -> OperationResult {
        let user = query!(
            "SELECT balance, debt FROM users WHERE username = $1",
            username.inner()
        )
        .fetch_optional(pool)
        .await?;

        match user {
            Some(user) => Ok(ServerResponse::UserInfo(User::new(
                username,
                UserBalance::new(user.balance),
                UserDebt::new(user.debt),
            ))),
            None => Ok(ServerResponse::UserError(UserError::InvalidUsername)),
        }
    }
}

/// User operation implementation
impl AppState {
    pub(crate) async fn login(&mut self, api_key: ApiKey, pool: &PgPool) -> OperationResult {
        self.assert_state(AppState::Unauthorized)?;
        let user_id = {
            let query = query!(
                "SELECT (user_id) FROM user_api_keys
                JOIN api_keys ON user_api_keys.key_id = api_keys.key_id
                WHERE key = $1",
                api_key.inner()
            )
            .fetch_optional(pool)
            .await?;
            match query {
                Some(id) => id.user_id,
                None => return Ok(ServerResponse::UserError(UserError::InvalidApiKey)),
            }
        };
        *self = AppState::Authorized { user_id };
        Ok(ServerResponse::Success)
    }

    pub(crate) async fn register(
        &mut self,
        username: Username,
        password: Password,
        pool: &PgPool,
    ) -> OperationResult {
        self.assert_state(AppState::Unauthorized)?;
        let salted_password =
            SaltedPassword::try_from(password).map_err(|_| StateError::PasswordHash)?;
        query!(
            "INSERT INTO users
            (username, pwhash_data)
            VALUES ($1, $2)",
            username.inner(),
            salted_password.as_ref(),
        )
        .execute(pool)
        .await?;
        Ok(ServerResponse::Success)
    }

    pub(crate) async fn generate_api_key(
        &self,
        username: Username,
        password: Password,
        pool: &PgPool,
    ) -> OperationResult {
        self.assert_state(AppState::Unauthorized)?;
        match self.validate_password(&username, &password, pool).await {
            Ok(None) => Ok(ServerResponse::UserError(UserError::InvalidPassword)),
            Ok(Some(user_id)) => self.create_api_key(user_id, pool).await,
            Err(_) => todo!(),
        }
    }

    async fn create_api_key(&self, user_id: i32, pool: &PgPool) -> OperationResult {
        let mut transaction = pool.begin().await?;
        let api_key = uuid::Uuid::new_v4();
        let api_key_id = query!(
            "INSERT INTO api_keys (key)
            VALUES ($1)
            RETURNING api_keys.key_id;",
            &api_key,
        )
        .fetch_one(&mut transaction)
        .await?
        .key_id;

        query!(
            "INSERT INTO user_api_keys (user_id, key_id)
            VALUES ($1, $2)",
            user_id,
            api_key_id
        )
        .execute(&mut transaction)
        .await?;

        transaction.commit().await?;

        Ok(ServerResponse::NewApiKey(api_key))
    }

    async fn validate_password(
        &self,
        username: &Username,
        password: &Password,
        pool: &PgPool,
    ) -> Result<Option<i32>, StateError> {
        let result = query!(
            "SELECT user_id, pwhash_data FROM users
            WHERE username = $1",
            username.inner()
        )
        .fetch_one(pool)
        .await?;

        let login_attempt = HashedPassword::from_slice(&result.pwhash_data)
            .map(|pw| pwhash_verify(&pw, password.as_bytes()))
            .unwrap();

        if login_attempt {
            Ok(Some(result.user_id))
        } else {
            Ok(None)
        }
    }
}

/// Market operation implementation
impl AppState {
    pub(crate) async fn buy(
        &self,
        symbol: StockSymbol,
        amount: usize,
        pool: &PgPool,
    ) -> OperationResult {
        let id = *match self {
            Self::Authorized { user_id } => user_id,
            _ => return Err(StateError::WrongState),
        };
        let user_balance = Self::get_user_balance(id, pool).await?;

        todo!()
    }

    pub(crate) async fn sell(
        &self,
        symbol: StockSymbol,
        amount: usize,
        pool: &PgPool,
    ) -> OperationResult {
        let id = *match self {
            Self::Authorized { user_id } => user_id,
            _ => return Err(StateError::WrongState),
        };
        let num_owned = Self::get_owned_stock_count(id, symbol, pool).await?;
        if num_owned < amount {
            return Ok(ServerResponse::UserError(UserError::NotEnoughOwnedStock(
                num_owned,
            )));
        }
        todo!()
    }

    async fn get_user_balance(user_id: i32, pool: &PgPool) -> Result<Decimal, StateError> {
        query!("SELECT balance FROM users where user_id = $1", user_id)
            .fetch_one(pool)
            .await
            .map(|record| record.balance)
            .map_err(StateError::from)
    }

    async fn get_owned_stock_count(
        user_id: i32,
        symbol: StockSymbol,
        pool: &PgPool,
    ) -> Result<usize, StateError> {
        query!(
            "SELECT amount FROM users_stocks
            JOIN stocks ON users_stocks.stock_id = stocks.stock_id
            WHERE user_id = $1 AND symbol = $2",
            user_id,
            symbol.inner()
        )
        .fetch_one(pool)
        .await
        .map(|record| record.amount as usize)
        .map_err(StateError::from)
    }
}

impl Default for AppState {
    fn default() -> Self {
        Self::new()
    }
}
